"""New VM Request Management plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import autocompletion, const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.request import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('new', short_help='Manage new virtual machine deployment requests')
@pass_context
def request_mgmt_new(ctx: Configuration):
    """Manage new virtual machine deployment requests."""
    pass


@request_mgmt_new.command('ls', short_help='list vm new requests')
@so.filter_opt
@so.all_opt
@so.count_opt
@so.page_opt
@so.sort_opt
@pass_context
def request_mgmt_new_ls(
    ctx: Configuration, filter_by, page, sort, show_all, count
):
    """List requests.

    Filter list in the following format <field_name> <operator>,<value>
    where operator is eq, ne, lt, le, gt, ge, like, in.
    For example: status=eq,PROCESSED

    vss-cli request new ls -f status=eq,PROCESSED

    Sort list in the following format <field_name>=<asc|desc>. For example:

    vss-cli request new ls -s created_on=desc
    """
    columns = ctx.columns or const.COLUMNS_REQUEST_NEW_MIN
    _LOGGING.debug(f'Columns {columns}')
    params = dict(expand=1, sort='created_on,desc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        _requests = ctx.get_new_requests(
            show_all=show_all, per_page=count, **params
        )

    output = format_output(ctx, _requests, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@request_mgmt_new.command('get', short_help='New vm request')
@click.argument(
    'rid',
    type=click.INT,
    required=True,
    shell_complete=autocompletion.new_requests,
)
@pass_context
def request_mgmt_new_get(ctx: Configuration, rid):
    """Get New VM Request info."""
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_new_request(rid)
    columns = ctx.columns or const.COLUMNS_REQUEST_NEW
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@request_mgmt_new.command('retry', short_help='Retry vm new request')
@click.argument('rid', type=click.INT, required=True)
@pass_context
def request_mgmt_new_retry(ctx: Configuration, rid):
    """Retry new VM request with status 'Error Processed'."""
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.retry_new_request(rid)
    columns = ctx.columns or const.COLUMNS_REQUEST_SUBMITTED
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))
    # wait for request
    if ctx.wait_for_requests:
        ctx.wait_for_request_to(obj)



"""Details for the auto-completion."""
import logging
import os
from typing import Callable, Dict, List, Optional, Tuple  # NOQA

import click
from click.shell_completion import CompletionItem
from requests.exceptions import HTTPError

from vss_cli import const
from vss_cli.config import Configuration
from vss_cli.exceptions import VssError

_LOGGING = logging.getLogger(__name__)


def _init_ctx(ctx: Configuration) -> None:
    """Initialize ctx."""
    # ctx is incomplete thus need to 'hack' around it
    # see bug https://github.com/pallets/click/issues/942
    ctx.client = Configuration(tk=os.environ.get('VSS_TOKEN'))
    ctx.client.endpoint = os.environ.get('VSS_ENDPOINT', None)
    ctx.client.username = os.environ.get('VSS_USER', None)
    ctx.client.password = os.environ.get('VSS_USER_PASS', None)
    ctx.client.timeout = int(
        os.environ.get('VSS_TIMEOUT', str(const.DEFAULT_TIMEOUT))
    )
    ctx.client.config_path = os.environ.get('VSS_CONFIG', const.DEFAULT_CONFIG)
    # fallback to load configuration
    ctx.client.load_config()


def to_completion_item(func: Callable):
    """Decorate function and converts to CI."""

    def wrapper_tci(*args, **kwargs):
        results = func(*args, **kwargs)
        return [CompletionItem(item[0], help=f'{item[1]}') for item in results]

    return wrapper_tci


def _autocomplete(
    f: Callable,
    incomplete: str,
    attrs: List[str],
    f_kwargs: Dict = None,
    sort_index: int = 0,
    complete_index: int = 0,
) -> List[Tuple[str, str]]:
    """Autocomplete main function."""
    try:
        response = f(**f_kwargs)
    except (HTTPError, VssError):
        response = []
    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            sub_attrs = []
            # start iterating in second item (index 1)
            for attr in attrs[1:]:
                sub_attrs.append(str(obj.get(attr, 'N/A')))
            # first item (index 0) is always the key
            r = (str(obj[attrs[0]]), ' - '.join(sub_attrs))
            completions.append(r)

        completions.sort(key=lambda x: x[sort_index])

        return [c for c in completions if incomplete in c[complete_index]]
    return completions


@to_completion_item
def table_formats(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Provide table Formats."""
    _init_ctx(ctx)

    completions = [
        ("plain", "Plain tables, no pseudo-graphics to draw lines"),
        ("simple", "Simple table with --- as header/footer (default)"),
        ("github", "Github flavored Markdown table"),
        ("grid", "Formatted as Emacs 'table.el' package"),
        ("fancy_grid", "Draws a fancy grid using box-drawing characters"),
        ("pipe", "PHP Markdown Extra"),
        ("orgtbl", "org-mode table"),
        ("jira", "Atlassian Jira Markup"),
        ("presto", "Formatted as PrestoDB cli"),
        ("psql", "Formatted as Postgres psql cli"),
        ("rst", "reStructuredText"),
        ("mediawiki", "Media Wiki as used in Wikpedia"),
        ("moinmoin", "MoinMain Wiki"),
        ("youtrack", "Youtrack format"),
        ("html", "HTML Markup"),
        ("pretty", "HTML escaping."),
        ("latex", "LaTeX markup, replacing special characters"),
        ("latex_raw", "LaTeX markup, no replacing of special characters"),
        (
            "latex_booktabs",
            "LaTex markup using spacing and style from `booktabs",
        ),
        ("textile", "Textile"),
        ("tsv", "Tab Separated Values"),
        ("csv", "Comma Separated Values"),
    ]

    completions.sort()

    return [c for c in completions if incomplete in c[0]]


@to_completion_item
def vm_templates(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM templates."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_templates,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


@to_completion_item
def virtual_machines(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VMs."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vms,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"show_all": True, "short": 1, "per_page": 2000},
    )


@to_completion_item
def vm_controller_scsi_sharing(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM controller SCSI types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_scsi_sharing,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


@to_completion_item
def vm_controller_scsi_types(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM controller SCSI types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_scsi_controllers,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


@to_completion_item
def vm_disk_backing_modes(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Disk Backing Modes."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_disk_backing_modes,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


@to_completion_item
def vm_disk_sharing(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Disk Backing Modes."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_disk_sharing,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


@to_completion_item
def domains(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Domains."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_domains,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


@to_completion_item
def folders(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM folders."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folders,
        incomplete,
        ['moref', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={
            "show_all": True,
            "sort": "path,desc",
            "per_page": 500,
            "short": 1,
        },
    )


@to_completion_item
def networks(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Networks."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_networks,
        incomplete,
        ['moref', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


@to_completion_item
def operating_systems(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM operating systems."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_os,
        incomplete,
        ['guest_id', 'full_name'],
        f_kwargs={"show_all": True, "sort": "guest_id,desc", "per_page": 500},
    )


@to_completion_item
def vss_services(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VSS Services."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vss_services,
        incomplete,
        ['id', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


@to_completion_item
def isos(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[CompletionItem]:
    """Autocomplete ISO images."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_isos,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


@to_completion_item
def clib_deployable_items(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete deployable items from content library."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_content_library_items,
        incomplete,
        attrs=['id', 'name', 'type'],
        sort_index=1,
        complete_index=1,
        f_kwargs={
            "show_all": True,
            "sort": "path,desc",
            "per_page": 500,
            "filter": "type,in,OVF,VM_TEMPLATE",
        },
    )


@to_completion_item
def vm_images(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM images."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_images,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


@to_completion_item
def floppies(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Floppies."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_floppies,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


@to_completion_item
def inventory_properties(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Inventory Properties."""
    _init_ctx(ctx)
    try:
        response = ctx.client.request('/inventory/options')
        if response:
            response = response.get('data')
    except (HTTPError, VssError):
        response = []

    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            completions.append((obj['key'], obj['value']))

        completions.sort()

        return [c for c in completions if incomplete in c[0]]
    return completions


@to_completion_item
def inventory_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Inventory Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_inventory_requests,
        incomplete,
        attrs=['id', 'created_on', 'name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def new_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete New VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_new_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def change_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Change VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_change_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name', 'attribute'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def export_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Export VM Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_export_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def folder_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Folder Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folder_requests,
        incomplete,
        attrs=['id', 'moref', 'action'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def image_sync_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Image Sync Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_image_sync_requests,
        incomplete,
        attrs=['id', 'type'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def vmdk_sync_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Image Sync Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vmdk_sync_requests,
        incomplete,
        attrs=['id', 'created_on'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def snapshot_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Snapshot Requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_snapshot_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def account_messages(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete Account Messages."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_user_messages,
        incomplete,
        attrs=['id', 'kind', 'subject'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def virtual_nic_types(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM NIC Types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_nic_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


@to_completion_item
def virtual_hw_types(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM Virtual Hardware Types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vmx_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


@to_completion_item
def vss_options(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VSS Options."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vss_options,
        incomplete,
        attrs=['option', 'description'],
        f_kwargs={"only_option": False},
    )


@to_completion_item
def groups(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete User Groups."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_groups,
        incomplete,
        attrs=['id', 'name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def vm_firmware(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM firmware types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_firmware_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


@to_completion_item
def vm_storage_type(
    ctx: Configuration,
    param: click.Option,
    incomplete: str,
) -> List[Tuple[str, str]]:
    """Autocomplete VM storage types."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_storage_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


@to_completion_item
def vm_snapshots(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Optional[Tuple[str, str]]]:
    """Autocomplete VM snapshots."""
    _init_ctx(ctx)
    args = ctx.params
    if 'source' in args:
        try:
            vm_id = args['source']
            return _autocomplete(
                ctx.client.get_vm_snapshots,
                incomplete,
                attrs=['id', 'name', 'description'],
                f_kwargs={'vm_id': vm_id},
            )
        except KeyError:
            return []
    return []


@to_completion_item
def retirement_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM retirement requests."""
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_retirement_requests,
        incomplete,
        attrs=['id', 'vm_moref', 'vm_name', 'retire_on'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


@to_completion_item
def vm_retirement_requests(
    ctx: Configuration, param: click.Option, incomplete: str
) -> List[Tuple[str, str]]:
    """Autocomplete VM retirement requests by VM."""
    _init_ctx(ctx)
    args = ctx.params
    if 'source' in args:
        try:
            vm_id = args['source']
            return _autocomplete(
                ctx.client.get_vm_retirement_requests,
                incomplete,
                attrs=['id', 'name', 'description'],
                f_kwargs={'vm_id': vm_id},
            )
        except IndexError:
            return []
    return []



"""VSS check vpn status."""
import logging
from typing import Dict, List

import requests

from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.vssconst import VSS_VPN_ENDPOINT

_LOGGING = logging.getLogger(__name__)


def check_vpn_status() -> List[Dict]:
    """Check vpn status."""
    icon = EMOJI_UNICODE.get(':question_mark:')
    result = []
    try:
        r = requests.get(VSS_VPN_ENDPOINT)
        _status = r.json()

        vpn_instances = _status.get('vpn')
        for instance in vpn_instances:
            status = instance.get('stat', 'unknown')
            iid = instance['id']
            if status in ['up']:
                status = 'operational'
                icon = EMOJI_UNICODE.get(':white_heavy_check_mark:')
            elif status in ['down']:
                icon = EMOJI_UNICODE.get(':cross_mark:')
            vpn_instance = {
                "name": f"ITS Private Cloud VPN {iid.upper()}",
                "status": status,
                "icon": icon,
            }
            result.append(vpn_instance)
    except Exception as ex:
        _LOGGING.warning(f'VSS VPN lookup failed: {ex}')
    return result


def check_stor_status():
    """TODO: Check stor status."""
    return



"""VSS Storage Management plugin for VSS CLI (vss-cli)."""
from datetime import timedelta
import logging
import os

import click

from vss_cli import const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.validators import validate_json_type

_LOGGING = logging.getLogger(__name__)


@click.group('stor', short_help='Manage your VSS storage account.')
@pass_context
def cli(ctx: Configuration):
    """Manage your VSS storage account."""
    with ctx.spinner(disable=ctx.debug) as spinner_cls:
        ctx.load_config(spinner_cls=spinner_cls)


@cli.command('la', short_help='launch ui')
@click.argument(
    'ui_type', type=click.Choice(['gui', 'admin', 's3api']), required=True
)
@click.option(
    '--show-cred/--no-show-cred',
    help='Show credentials or not',
    is_flag=True,
)
@pass_context
def stor_launch(ctx: Configuration, ui_type, show_cred):
    """Launch web ui."""
    ctx.get_vskey_stor()
    opts = {
        'gui': ctx.vskey_stor_s3_gui,
        's3api': ctx.vskey_stor_s3api,
        'admin': ctx.s3_server,
    }
    url = opts[ui_type]
    click.echo(f'Launching {EMOJI_UNICODE[":globe_showing_Americas:"]}: {url}')
    if ui_type in ['gui', 's3api']:
        if show_cred:
            click.echo(
                f'username: {ctx.vskey_stor_s3_ak}\n'
                f'password: {ctx.vskey_stor_s3_sk}\n'
            )
    click.launch(
        url,
    )


@cli.command('ls', short_help='list remote dir contents')
@click.option(
    '--bucket', type=click.STRING, help='Bucket to list.', default='ut-vss'
)
@click.argument('remote_path', type=click.STRING, default="/")
@pass_context
def stor_ls(ctx: Configuration, bucket, remote_path):
    """List contents of your VSKEY-STOR Space."""
    columns = ctx.columns or const.COLUMNS_STOR
    ctx.get_vskey_stor()
    with ctx.spinner(disable=ctx.debug):
        obj_gen = ctx.vskey_stor.list_objects(
            bucket, remote_path, recursive=True
        )
        obj = [o.object_name for o in obj_gen]
    click.echo(format_output(ctx, obj, columns=columns))


@cli.command('sh', short_help='share with pre-signed link')
@click.option(
    '--bucket',
    type=click.STRING,
    help='Bucket where file is stored.',
    default='ut-vss',
)
@click.option(
    '--expires', type=click.INT, help='Expiry time in hours.', default=2
)
@click.option('--launch', is_flag=True, help='Launch with default handler.')
@click.argument('remote_path', type=click.STRING, required=True)
@pass_context
def stor_share(ctx, remote_path, bucket, expires, launch):
    """Get a pre-signed link to download object."""
    ctx.get_vskey_stor()
    url = ctx.vskey_stor.get_presigned_url(
        "GET",
        bucket_name=bucket,
        object_name=remote_path,
        expires=timedelta(hours=expires),
    )
    obj = {'url': url}
    columns = ctx.columns or const.COLUMNS_STOR_SHARE
    click.echo(format_output(ctx, [obj], columns=columns, single=True))
    if launch:
        click.echo(
            f'Launching {EMOJI_UNICODE[":globe_showing_Americas:"]}: {url}'
        )
        click.launch(url)


@cli.command('get', short_help='get info')
@click.option(
    '--bucket', type=click.STRING, help='Bucket to list.', default='ut-vss'
)
@click.argument('remote_path', type=click.STRING, required=True)
@pass_context
def stor_get(ctx, remote_path, bucket):
    """Get file info."""
    ctx.get_vskey_stor()
    columns = ctx.columns or const.COLUMNS_STOR_INFO
    obj_gen = ctx.vskey_stor.list_objects(
        bucket_name=bucket, prefix=remote_path
    )
    objs = [
        {
            'name': o.object_name,
            'bucket_name': o.bucket_name,
            'last_modified': o.last_modified,
            'size': o.size,
        }
        for o in obj_gen
    ]
    click.echo(format_output(ctx, objs, columns=columns, single=True))


@cli.command('dl', short_help='download file')
@click.argument('remote_path', type=click.STRING, required=True)
@click.option(
    '--bucket', type=click.STRING, help='Bucket to list.', default='ut-vss'
)
@click.option(
    '-d',
    '--dir',
    type=click.STRING,
    help='Local target directory',
    default=os.getcwd(),
)
@click.option('-n', '--name', type=click.STRING, help='Local target name')
@pass_context
def stor_dl(ctx: Configuration, remote_path, dir, name, bucket):
    """Download remote file."""
    ctx.get_vskey_stor()
    local_dir = os.path.expanduser(dir) or os.getcwd()
    local_name = name or os.path.basename(remote_path)
    local_path = os.path.join(local_dir, local_name)
    # check if remote path exists
    ctx.log(
        f'Download {remote_path} to {local_path} in '
        f'progress {EMOJI_UNICODE.get(":fast_down_button:")} '
    )
    with ctx.spinner(disable=ctx.debug):
        ctx.vskey_stor.fget_object(
            bucket_name=bucket, object_name=remote_path, file_path=local_path
        )
    ctx.log(
        f'Download complete to {local_path} '
        f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}'
    )


@cli.command('ul', short_help='upload file')
@click.argument('file_path', type=click.Path(exists=True), required=True)
@click.option(
    '--bucket', type=click.STRING, help='Bucket to list.', default='ut-vss'
)
@click.option(
    '-d',
    '--dir',
    type=click.STRING,
    help='Remote target directory',
    default='/',
)
@click.option(
    '-n',
    '--name',
    type=click.STRING,
    help='Remote target name',
    required=False,
)
@click.option(
    '-m',
    '--metadata',
    help='Key-Value JSON metadata.',
    callback=validate_json_type,
    type=click.STRING,
    required=False,
)
@pass_context
def stor_ul(ctx: Configuration, file_path, name, dir, bucket, metadata):
    """Upload file to your VSKEY-STOR space.

    This command is useful when a required ISO is
    not available in the VSS central repository and needs to be
    mounted to a virtual machine.
    """
    ctx.get_vskey_stor()
    file_name = name or os.path.basename(file_path)
    remote_base = dir
    # upload
    remote_path = os.path.join(remote_base, file_name)
    ctx.log(
        f'Upload {file_path} to {remote_path} '
        f'in progress {EMOJI_UNICODE.get(":fast_up_button:")}'
    )
    with ctx.spinner(disable=ctx.debug):
        kwargs = dict(
            bucket_name=bucket, object_name=file_name, file_path=file_path
        )
        if metadata is not None:
            kwargs['metadata'] = metadata
        result = ctx.vskey_stor.fput_object(**kwargs)
    ctx.echo(
        f'Upload complete to {result.object_name}, etag={result.etag}, '
        f'version-id={result.version_id} '
        f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}'
    )



"""OVF plugin for VSS CLI (vss-cli)."""
import logging
from pathlib import Path

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output, raw_format_output

_LOGGING = logging.getLogger(__name__)


@click.group('ovf', short_help='OVF Tool')
@pass_context
def cli(ctx: Configuration):
    """Manage OVF and OVA."""
    ctx.set_defaults()


@cli.group(
    'get',
    short_help='Load and analyze OVA or OVF.',
    invoke_without_command=True,
)
@click.argument('file_path', type=click.Path(exists=True), required=True)
@pass_context
def get_ovf(ctx: Configuration, file_path):
    """Load and analyze OVA or OVF."""
    ctx.ova_or_ovf = file_path
    obj = ctx.parse_ova_or_ovf(file_path)
    ctx.ovf_dict = obj
    if click.get_current_context().invoked_subcommand is None:
        columns = ctx.columns or const.COLUMNS_OVF
        click.echo(format_output(ctx, [obj], columns=columns, single=True))


@get_ovf.command('property-params', short_help='Get Property parameters.')
@so.page_opt
@pass_context
def get_ovf_property_params(ctx: Configuration, page):
    """Get OVF property params (if any)."""
    objs = ctx.ovf_dict.get('PropertyParams', [])
    # including additional attributes?
    columns = ctx.columns or const.COLUMNS_OVF_PP
    # format output
    output = format_output(ctx, objs, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@get_ovf.command(
    'deployment-params', short_help='Get Deployment Option parameters'
)
@so.page_opt
@pass_context
def get_ovf_deployment_params(ctx: Configuration, page):
    """Get OVF deployment params (if any)."""
    objs = ctx.ovf_dict.get('DeploymentOptionParams', [])
    # including additional attributes?
    columns = ctx.columns or const.COLUMNS_OVF_DP
    # format output
    output = format_output(ctx, objs, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@get_ovf.command('params-spec', short_help='Generate additional-params spec')
@click.option(
    '--edit/--no-edit',
    is_flag=True,
    required=False,
    help='Edit before writing',
)
@pass_context
def get_ovf_additional_params_spec(ctx: Configuration, edit):
    """Generate additional params file spec."""
    new_raw = None
    if ctx.output in ['auto', 'table']:
        _LOGGING.warning(f'Input set to {ctx.output}. Falling back to yaml')
        ctx.output = 'yaml'
    source_name = Path(ctx.ova_or_ovf).stem
    deployment_params_comment = ''
    property_params_comment = ''
    f_name = f'{source_name}-additional-params.{ctx.output}'
    obj = ctx.ovf_dict

    n_obj = {}
    if obj.get('PropertyParams'):
        property_params = {
            x.get('key'): x.get('default')
            for x in obj.get('PropertyParams', [{}])
        }
        property_params_comment = '\n'.join(
            [
                f'# {x["key"]} ({x["type"]}): {x["description"]}'.replace(
                    '\n', ''
                )
                for x in obj.get('PropertyParams', [{}])
            ]
        )
        property_params_comment = (
            f'# PropertyParams:\n{property_params_comment}'
        )
        n_obj['PropertyParams'] = property_params
    if obj.get('DeploymentOptionParams'):
        deployment_params_comment = '\n'.join(
            [
                f'# {x["id"]}: {x["description"]}'
                for x in obj.get('DeploymentOptionParams', [{}])
            ]
        )
        deployment_params_comment = (
            f'# DeploymentOptionParams:\n{deployment_params_comment}'
        )
        deploy_params = {'selected_key': ''}
        n_obj['DeploymentOptionParams'] = deploy_params

    if edit:
        obj_raw = raw_format_output(
            ctx.output, n_obj, ctx.yaml(), highlighted=False
        )
        obj_raw = '\n'.join(
            [obj_raw, deployment_params_comment, property_params_comment]
        )
        new_raw = click.edit(obj_raw, extension=f'.{ctx.output}')

    if new_raw:
        n_obj = ctx.yaml_load(new_raw)
    # writing objects
    with open(f_name, 'w') as fp:
        ctx.yaml_dump_stream(n_obj, stream=fp)
    ctx.echo(f'Written to {f_name}')



"""Vm Snapshot Request Management plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import const, rel_opts as so
import vss_cli.autocompletion as autocompletion
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.request import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('snapshot', short_help='Manage virtual machine snapshot requests')
@pass_context
def snapshot(ctx: Configuration):
    """Manage VM snapshot requests.

    Creating, deleting and reverting virtual machine
    snapshots will produce a virtual machine snapshot request.
    """


@snapshot.command('ls', short_help='list snapshot requests')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def snapshot_ls(ctx: Configuration, filter_by, page, sort, show_all, count):
    """List requests.

    Filter list in the following format <field_name>=<operator>,<value>
    where operator is eq, ne, lt, le, gt, ge, like, in.
    For example: status=eq,PROCESSED

    vss-cli request snapshot ls -f status=eq,PROCESSED

    Sort list in the following format <field_name>=<asc|desc>. For example:

    vss-cli request snapshot ls -s created_on=desc
    """
    columns = ctx.columns or const.COLUMNS_REQUEST_SNAP_MIN
    params = dict(expand=1, sort='created_on,desc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        _requests = ctx.get_snapshot_requests(
            show_all=show_all, per_page=count, **params
        )

    output = format_output(ctx, _requests, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        ctx.echo(output)


@snapshot.command('get', help='Snapshot request')
@click.argument(
    'rid',
    type=click.INT,
    required=True,
    shell_complete=autocompletion.snapshot_requests,
)
@pass_context
def snapshot_get(ctx, rid):
    """Get VM snapshot request info."""
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_snapshot_request(rid)
    columns = ctx.columns or const.COLUMNS_REQUEST
    if not ctx.columns:
        columns.extend(const.COLUMNS_REQUEST_SNAP)
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))


@snapshot.group('set', help='Update snapshot request')
@click.argument(
    'rid',
    type=click.INT,
    required=True,
    shell_complete=autocompletion.snapshot_requests,
)
@pass_context
def snapshot_set(ctx: Configuration, rid):
    """Set snapshot attribute."""
    ctx.rid = rid


@snapshot_set.command('duration')
@click.option(
    '-l',
    '--lifetime',
    type=click.IntRange(1, 72),
    help='Number of hours the snapshot will live.',
    required=True,
)
@pass_context
def snapshot_set_duration(ctx: Configuration, lifetime):
    """Extend snapshot lifetime."""
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.extend_snapshot_request(ctx.rid, lifetime)
    columns = ctx.columns or const.COLUMNS_REQUEST
    if not ctx.columns:
        columns.extend(const.COLUMNS_REQUEST_SNAP)
    ctx.echo(format_output(ctx, [obj], columns=columns, single=True))
